@@ -48,6 +48,7 @@ from async_timeout import timeout
 import attr
 from cachetools import cached, LRUCache
 import snappy
+import time
 import zmq, zmq.asyncio
 
 from ai.backend.common import msgpack, redis
@@ -99,7 +100,7 @@ from ai.backend.common.events import (
     SessionFailureEvent,
     SessionSuccessEvent,
 )
-from ai.backend.common.utils import current_loop
+from ai.backend.common.utils import cancel_tasks, current_loop
 from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext
 from ai.backend.common.service_ports import parse_service_ports
 from . import __version__ as VERSION
@@ -208,6 +209,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
     error_monitor: ErrorPluginContext  # unused in favor of produce_error_event()
 
     _pending_creation_tasks: Dict[KernelId, Set[asyncio.Task]]
+    _ongoing_exec_batch_tasks: weakref.WeakSet[asyncio.Task]
     _ongoing_destruction_tasks: weakref.WeakValueDictionary[KernelId, asyncio.Task]
 
     def __init__(
@@ -241,6 +243,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         self.error_monitor = error_monitor
         self._rx_distro = re.compile(r"\.([a-z-]+\d+\.\d+)\.")
         self._pending_creation_tasks = defaultdict(set)
+        self._ongoing_exec_batch_tasks = weakref.WeakSet()
         self._ongoing_destruction_tasks = weakref.WeakValueDictionary()
 
     async def __ainit__(self) -> None:
@@ -249,35 +252,16 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         It must call this super method in an appropriate order, only once.
         """
         self.resource_lock = asyncio.Lock()
+        self.registry_lock = asyncio.Lock()
         self.container_lifecycle_queue = asyncio.Queue()
 
-        async def connect_redis_event_db() -> aioredis.Redis:
-            return await redis.connect_with_retries(
-                self.local_config['redis']['addr'].as_sockaddr(),
-                db=4,  # REDIS_STREAM_DB in gateway.defs
-                password=(self.local_config['redis']['password']
-                        if self.local_config['redis']['password'] else None),
-                encoding=None,
-            )
-
         self.event_producer = await EventProducer.new(
-            connect_redis_event_db,
+            self.local_config['redis'],
+            db=4,
             log_events=self.local_config['debug']['log-events'],
         )
-        self.redis_stream_pool = await redis.connect_with_retries(
-            self.local_config['redis']['addr'].as_sockaddr(),
-            db=4,  # REDIS_STREAM_DB in backend.ai-manager
-            password=(self.local_config['redis']['password']
-                      if self.local_config['redis']['password'] else None),
-            encoding=None,
-        )
-        self.redis_stat_pool = await redis.connect_with_retries(
-            self.local_config['redis']['addr'].as_sockaddr(),
-            db=0,  # REDIS_STAT_DB in backend.ai-manager
-            password=(self.local_config['redis']['password']
-                      if self.local_config['redis']['password'] else None),
-            encoding='utf8',
-        )
+        self.redis_stream_pool = redis.get_redis_object(self.local_config['redis'], db=4)
+        self.redis_stat_pool = redis.get_redis_object(self.local_config['redis'], db=0)
 
         ipc_base_path.mkdir(parents=True, exist_ok=True)
         self.zmq_ctx = zmq.asyncio.Context()
@@ -305,6 +289,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         self.timer_tasks.append(aiotools.create_timer(self.sync_container_lifecycles, 10.0))
 
         loop = current_loop()
+        self.last_registry_written_time = time.monotonic()
         self.container_lifecycle_handler = loop.create_task(self.process_lifecycle_events())
 
         # Notify the gateway.
@@ -315,19 +300,20 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         An implementation of AbstractAgent would define its own ``shutdown()`` method.
         It must call this super method in an appropriate order, only once.
         """
-        # Close all pending kernel runners.
-        for kernel_obj in self.kernel_registry.values():
-            if kernel_obj.runner is not None:
-                await kernel_obj.runner.close()
-            await kernel_obj.close()
-        if stop_signal == signal.SIGTERM:
-            await self.clean_all_kernels(blocking=True)
+        await cancel_tasks(self._ongoing_exec_batch_tasks)
+
+        async with self.registry_lock:
+            # Close all pending kernel runners.
+            for kernel_obj in self.kernel_registry.values():
+                if kernel_obj.runner is not None:
+                    await kernel_obj.runner.close()
+                await kernel_obj.close()
+            if stop_signal == signal.SIGTERM:
+                await self.clean_all_kernels(blocking=True)
 
         # Stop timers.
-        for task in self.timer_tasks:
-            task.cancel()
-        timer_cancel_results = await asyncio.gather(*self.timer_tasks, return_exceptions=True)
-        for result in timer_cancel_results:
+        cancel_results = await cancel_tasks(self.timer_tasks)
+        for result in cancel_results:
             if isinstance(result, Exception):
                 log.error('timer cancellation error: {}', result)
 
@@ -338,12 +324,10 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         # Notify the gateway.
         await self.produce_event(AgentTerminatedEvent(reason="shutdown"))
 
-        # Close Redis connection pools.
+        # Shut down the event dispatcher and Redis connection pools.
         await self.event_producer.close()
-        self.redis_stream_pool.close()
-        await self.redis_stream_pool.wait_closed()
-        self.redis_stat_pool.close()
-        await self.redis_stat_pool.wait_closed()
+        await self.redis_stream_pool.close()
+        await self.redis_stat_pool.close()
 
         self.zmq_ctx.term()
 
@@ -433,9 +417,11 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                     while chunk_length >= chunk_size:
                         cb = chunk_buffer.getbuffer()
                         stored_chunk = bytes(cb[:chunk_size])
-                        await redis.execute_with_retries(
-                            lambda: self.redis_stream_pool.rpush(
-                                log_key, stored_chunk)
+                        await redis.execute(
+                            self.redis_stream_pool,
+                            lambda r: r.rpush(
+                                log_key, stored_chunk,
+                            ),
                         )
                         remaining = cb[chunk_size:]
                         chunk_length = len(remaining)
@@ -446,9 +432,11 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                         chunk_buffer = next_chunk_buffer
             assert chunk_length < chunk_size
             if chunk_length > 0:
-                await redis.execute_with_retries(
-                    lambda: self.redis_stream_pool.rpush(
-                        log_key, chunk_buffer.getvalue())
+                await redis.execute(
+                    self.redis_stream_pool,
+                    lambda r: r.rpush(
+                        log_key, chunk_buffer.getvalue(),
+                    ),
                 )
         finally:
             chunk_buffer.close()
@@ -456,8 +444,9 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         # This is just a safety measure to prevent memory leak in Redis
         # for cases when the event delivery has failed or processing
         # the log data has failed.
-        await redis.execute_with_retries(
-            lambda: self.redis_stream_pool.expire(log_key, 3600.0)
+        await redis.execute(
+            self.redis_stream_pool,
+            lambda r: r.expire(log_key, 3600),
         )
         await self.produce_event(DoSyncKernelLogsEvent(kernel_id, container_id))
 
@@ -478,12 +467,13 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         try:
             updated_kernel_ids = []
             container_ids = []
-            for kernel_id, kernel_obj in [*self.kernel_registry.items()]:
-                if not kernel_obj.stats_enabled:
-                    continue
-                updated_kernel_ids.append(kernel_id)
-                container_ids.append(kernel_obj['container_id'])
-            await self.stat_ctx.collect_container_stat(container_ids)
+            async with self.registry_lock:
+                for kernel_id, kernel_obj in [*self.kernel_registry.items()]:
+                    if not kernel_obj.stats_enabled:
+                        continue
+                    updated_kernel_ids.append(kernel_id)
+                    container_ids.append(kernel_obj['container_id'])
+                await self.stat_ctx.collect_container_stat(container_ids)
             # Let the manager store the statistics in the persistent database.
             if updated_kernel_ids:
                 await self.produce_event(DoSyncKernelStatsEvent(updated_kernel_ids))
@@ -494,54 +484,55 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             await self.produce_error_event()
 
     async def _handle_start_event(self, ev: ContainerLifecycleEvent) -> None:
-        kernel_obj = self.kernel_registry.get(ev.kernel_id)
-        if kernel_obj is not None:
-            kernel_obj.stats_enabled = True
+        async with self.registry_lock:
+            kernel_obj = self.kernel_registry.get(ev.kernel_id)
+            if kernel_obj is not None:
+                kernel_obj.stats_enabled = True
 
     async def _handle_destroy_event(self, ev: ContainerLifecycleEvent) -> None:
-        result = None
         try:
             current_task = asyncio.current_task()
             assert current_task is not None
             if ev.kernel_id not in self._ongoing_destruction_tasks:
                 self._ongoing_destruction_tasks[ev.kernel_id] = current_task
             self.terminating_kernels.add(ev.kernel_id)
-            kernel_obj = self.kernel_registry.get(ev.kernel_id)
-            if kernel_obj is None:
-                log.warning('destroy_kernel(k:{0}) kernel missing (already dead?)',
-                            ev.kernel_id)
-                if ev.container_id is None:
-                    await self.rescan_resource_usage()
-                    if not ev.suppress_events:
-                        await self.produce_event(
-                            KernelTerminatedEvent(ev.kernel_id, "already-terminated")
+            async with self.registry_lock:
+                kernel_obj = self.kernel_registry.get(ev.kernel_id)
+                if kernel_obj is None:
+                    log.warning('destroy_kernel(k:{0}) kernel missing (already dead?)',
+                                ev.kernel_id)
+                    if ev.container_id is None:
+                        await self.rescan_resource_usage()
+                        if not ev.suppress_events:
+                            await self.produce_event(
+                                KernelTerminatedEvent(ev.kernel_id, "already-terminated"),
+                            )
+                        if ev.done_event is not None:
+                            ev.done_event.set()
+                        return
+                    else:
+                        await self.container_lifecycle_queue.put(
+                            ContainerLifecycleEvent(
+                                ev.kernel_id,
+                                ev.container_id,
+                                LifecycleEvent.CLEAN,
+                                ev.reason,
+                                suppress_events=ev.suppress_events,
+                                done_event=ev.done_event,
+                            ),
                         )
-                    return
                 else:
-                    await self.container_lifecycle_queue.put(
-                        ContainerLifecycleEvent(
-                            ev.kernel_id,
-                            ev.container_id,
-                            LifecycleEvent.CLEAN,
-                            ev.reason,
-                            suppress_events=ev.suppress_events,
-                        )
-                    )
-            else:
-                kernel_obj.stats_enabled = False
-                kernel_obj.termination_reason = ev.reason
-                if kernel_obj.runner is not None:
-                    await kernel_obj.runner.close()
-            result = await self.destroy_kernel(ev.kernel_id, ev.container_id)
+                    kernel_obj.stats_enabled = False
+                    kernel_obj.termination_reason = ev.reason
+                    if kernel_obj.runner is not None:
+                        await kernel_obj.runner.close()
+                    kernel_obj.clean_event = ev.done_event
+                await self.destroy_kernel(ev.kernel_id, ev.container_id)
         except asyncio.CancelledError:
             pass
         except Exception:
             log.exception('unhandled exception while processing DESTROY event')
             await self.produce_error_event()
-        finally:
-            if ev.done_event is not None:
-                ev.done_event.set()
-                setattr(ev.done_event, '_result', result)
 
     async def _handle_clean_event(self, ev: ContainerLifecycleEvent) -> None:
         destruction_task = self._ongoing_destruction_tasks.get(ev.kernel_id, None)
@@ -549,59 +540,65 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             # let the destruction task finish first
             await destruction_task
             del destruction_task
-        result = None
-        try:
-            kernel_obj = self.kernel_registry.get(ev.kernel_id)
-            if kernel_obj is not None and kernel_obj.runner is not None:
-                await kernel_obj.runner.close()
-            result = await self.clean_kernel(
-                ev.kernel_id,
-                ev.container_id,
-                ev.kernel_id in self.restarting_kernels,
-            )
-        except Exception:
-            await self.produce_error_event()
-        finally:
+        async with self.registry_lock:
             try:
                 kernel_obj = self.kernel_registry.get(ev.kernel_id)
-                if kernel_obj is not None:
-                    # Restore used ports to the port pool.
-                    port_range = self.local_config['container']['port-range']
-                    # Exclude out-of-range ports, because when the agent restarts
-                    # with a different port range, existing containers' host ports
-                    # may not belong to the new port range.
-                    restored_ports = [*filter(
-                        lambda p: port_range[0] <= p <= port_range[1],
-                        kernel_obj['host_ports']
-                    )]
-                    self.port_pool.update(restored_ports)
-                    await kernel_obj.close()
-                    # Notify cleanup waiters.
-                    if kernel_obj.clean_event is not None:
-                        kernel_obj.clean_event.set()
-                    # Forget.
-                    self.kernel_registry.pop(ev.kernel_id, None)
+                if kernel_obj is not None and kernel_obj.runner is not None:
+                    await kernel_obj.runner.close()
+                await self.clean_kernel(
+                    ev.kernel_id,
+                    ev.container_id,
+                    ev.kernel_id in self.restarting_kernels,
+                )
+            except Exception:
+                await self.produce_error_event()
             finally:
-                if ev.done_event is not None:
-                    ev.done_event.set()
-                    setattr(ev.done_event, '_result', result)
-                if restart_tracker := self.restarting_kernels.get(ev.kernel_id, None):
-                    restart_tracker.destroy_event.set()
+                if ev.kernel_id in self.restarting_kernels:
+                    # Don't forget as we are restarting it.
+                    kernel_obj = self.kernel_registry.get(ev.kernel_id, None)
                 else:
-                    await self.rescan_resource_usage()
-                    if not ev.suppress_events:
-                        await self.produce_event(
-                            KernelTerminatedEvent(ev.kernel_id, ev.reason)
-                        )
-                self.terminating_kernels.discard(ev.kernel_id)
+                    # Forget as we are done with this kernel.
+                    kernel_obj = self.kernel_registry.pop(ev.kernel_id, None)
+                try:
+                    if kernel_obj is not None:
+                        # Restore used ports to the port pool.
+                        port_range = self.local_config['container']['port-range']
+                        # Exclude out-of-range ports, because when the agent restarts
+                        # with a different port range, existing containers' host ports
+                        # may not belong to the new port range.
+                        restored_ports = [*filter(
+                            lambda p: port_range[0] <= p <= port_range[1],
+                            kernel_obj['host_ports'],
+                        )]
+                        self.port_pool.update(restored_ports)
+                        await kernel_obj.close()
+                finally:
+                    self.terminating_kernels.discard(ev.kernel_id)
+                    if restart_tracker := self.restarting_kernels.get(ev.kernel_id, None):
+                        restart_tracker.destroy_event.set()
+                    else:
+                        await self.rescan_resource_usage()
+                        if not ev.suppress_events:
+                            await self.produce_event(
+                                KernelTerminatedEvent(ev.kernel_id, ev.reason),
+                            )
+                    # Notify cleanup waiters after all state updates.
+                    if kernel_obj is not None and kernel_obj.clean_event is not None:
+                        kernel_obj.clean_event.set()
+                    if ev.done_event is not None:
+                        ev.done_event.set()
 
     async def process_lifecycle_events(self) -> None:
         async with aiotools.TaskGroup() as tg:
             while True:
                 ev = await self.container_lifecycle_queue.get()
-                if isinstance(ev, Sentinel):
+                now = time.monotonic()
+                if now > self.last_registry_written_time + 60 or isinstance(ev, Sentinel):
+                    self.last_registry_written_time = now
                     with open(ipc_base_path / f'last_registry.{self.local_instance_id}.dat', 'wb') as f:
                         pickle.dump(self.kernel_registry, f)
+                    log.debug(f'saved last_registry.{self.local_instance_id}.dat')
+                if isinstance(ev, Sentinel):
                     return
                 # attr currently does not support customizing getstate/setstate dunder methods
                 # until the next release.
@@ -632,24 +629,40 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         container_id: ContainerId = None,
         exit_code: int = None,
         done_event: asyncio.Event = None,
-        clean_event: asyncio.Event = None,
         suppress_events: bool = False,
     ) -> None:
         try:
             kernel_obj = self.kernel_registry[kernel_id]
             if kernel_obj.termination_reason:
                 reason = kernel_obj.termination_reason
-            if kernel_obj.clean_event is not None:
-                # This should not happen!
-                log.warning('overwriting kernel_obj.clean_event (k:{})', kernel_id)
-            kernel_obj.clean_event = clean_event
-            if container_id is not None and container_id != kernel_obj['container_id']:
-                # This should not happen!
-                log.warning('container id mismatch for kernel_obj (k:{}, c:{}) with event (c:{})',
-                            kernel_id, kernel_obj['container_id'], container_id)
+            if container_id is not None:
+                if event == LifecycleEvent.START:
+                    # Update the container ID (for restarted kernels).
+                    # This will be overwritten by create_kernel() soon, but
+                    # updating here improves consistency of kernel_id to container_id
+                    # mapping earlier.
+                    kernel_obj['container_id'] = container_id
+                elif container_id != kernel_obj['container_id']:
+                    # This should not happen!
+                    log.warning(
+                        "container id mismatch for kernel_obj (k:{}, c:{}) with event (e:{}, c:{})",
+                        kernel_id, kernel_obj['container_id'],
+                        event.name, container_id,
+                    )
             container_id = kernel_obj['container_id']
         except KeyError:
-            pass
+            if event == LifecycleEvent.START:
+                # When creating a new kernel, the kernel_registry is not populated yet
+                # during creation of actual containers.
+                # The Docker daemon may publish the container creation event before
+                # returning the API and our async handlers may deliver the event earlier.
+                # In such cases, it is safe to ignore the missing kernel_regisry item.
+                pass
+            else:
+                log.warning(
+                    "injecting lifecycle event (e:{}) for unknown kernel (k:{})",
+                    event.name, kernel_id,
+                )
         await self.container_lifecycle_queue.put(
             ContainerLifecycleEvent(
                 kernel_id,
@@ -754,7 +767,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                 kernel_id,
                 LifecycleEvent.DESTROY,
                 'agent-termination',
-                clean_event=clean_events[kernel_id] if blocking else None,
+                done_event=clean_events[kernel_id] if blocking else None,
             )
         if blocking:
             waiters = [clean_events[kernel_id].wait()
@@ -1346,9 +1359,10 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         # adn specify the container mounts.
 
         # Mount backend-specific intrinsic mounts (e.g., scratch directories)
-        resource_spec.mounts.extend(
-            await self.create_kernel__get_intrinsic_mounts(ctx)
-        )
+        if not restarting:
+            resource_spec.mounts.extend(
+                await self.create_kernel__get_intrinsic_mounts(ctx)
+            )
 
         # Realize ComputeDevice (including accelerators) allocations.
         slots = resource_spec.slots
@@ -1382,21 +1396,23 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                         raise
 
         # Prepare scratch spaces and dotfiles inside it.
-        await self.create_kernel__prepare_scratch(ctx)
+        if not restarting:
+            await self.create_kernel__prepare_scratch(ctx)
 
         # Prepare networking.
         await self.create_kernel__apply_network(ctx, cluster_info)
         await self.create_kernel__install_ssh_keypair(ctx, cluster_info)
 
         # Mount vfolders and krunner stuffs.
-        await self._create_kernel__mount_vfolders(ctx, kernel_config['mounts'], resource_spec)
-        await self._create_kernel__mount_krunner(ctx, resource_spec, environ)
+        if not restarting:
+            await self._create_kernel__mount_vfolders(ctx, kernel_config['mounts'], resource_spec)
+            await self._create_kernel__mount_krunner(ctx, resource_spec, environ)
 
         # Inject Backend.AI-intrinsic env-variables for libbaihook and gosu
         label_envs_corecount = image_labels.get('ai.backend.envs.corecount', '')
         envs_corecount = label_envs_corecount.split(',') if label_envs_corecount else []
         cpu_core_count = len(resource_spec.allocations[DeviceName('cpu')][SlotName('cpu')])
-        environ.update({k: str(cpu_core_count) for k in envs_corecount})
+        environ.update({k: str(cpu_core_count) for k in envs_corecount if k not in environ})
 
         # Realize mounts.
         await self.create_kernel__process_mounts(
@@ -1527,6 +1543,13 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             KernelStartedEvent(kernel_id, creation_id)
         )
 
+        if kernel_config['session_type'] == 'batch' and kernel_config['cluster_role'] == 'main':
+            self._ongoing_exec_batch_tasks.add(
+                asyncio.create_task(
+                    self.execute_batch(kernel_id, kernel_config['startup_command'] or "")
+                )
+            )
+
         # The startup command for the batch-type sessions will be executed by the manager
         # upon firing of the "session_started" event.
 
@@ -1690,7 +1713,6 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                 )
                 raise
             else:
-                # tracker.destroy_event.clear()
                 try:
                     await self.create_kernel(
                         creation_id,
