@@ -108,6 +108,9 @@ class DockerKernel(AbstractKernel):
         })
         return result
 
+    async def shutdown_service(self, service: str):
+        await self.runner.feed_shutdown_service(service)
+
     async def get_service_apps(self):
         result = await self.runner.feed_service_apps()
         return result
@@ -142,7 +145,7 @@ class DockerKernel(AbstractKernel):
         except ValueError:
             raise PermissionError('You cannot download files outside /home/work')
         try:
-            with await container.get_archive(abspath) as tarobj:
+            with await container.get_archive(str(abspath)) as tarobj:
                 tarobj.fileobj.seek(0, 2)
                 fsize = tarobj.fileobj.tell()
                 if fsize > 1048576:
