@@ -31,9 +31,10 @@ from async_timeout import timeout
 import zmq, zmq.asyncio
 
 from ai.backend.common import msgpack
+from ai.backend.common.asyncio import current_loop
 from ai.backend.common.docker import ImageRef
+from ai.backend.common.enum_extension import StringSetFlag
 from ai.backend.common.types import aobject, KernelId
-from ai.backend.common.utils import current_loop, StringSetFlag
 from ai.backend.common.logging import BraceStyleAdapter
 from .resources import KernelResourceSpec
 
@@ -897,4 +898,5 @@ def match_distro_data(data: Mapping[str, Any], distro: str) -> Tuple[str, Any]:
         for distro_key, value in match_list:
             if distro_key == distro:
                 return (distro_key, value)
+        return match_list[0]  # fallback to the latest of its kind
     raise RuntimeError('no match found', distro)
