@@ -11,13 +11,18 @@ import logging
 import sys
 import time
 from typing import (
-    Optional,
     Callable,
-    List, Tuple,
-    Dict, Mapping, MutableMapping,
+    Dict,
     FrozenSet,
+    List,
+    Mapping,
+    MutableMapping,
+    Optional,
+    Sequence,
+    Tuple,
     TYPE_CHECKING,
 )
+import aioredis
 
 import attr
 
@@ -238,7 +243,7 @@ class StatContext:
     kernel_metrics: MutableMapping[KernelId, MutableMapping[MetricKey, Metric]]
 
     def __init__(self, agent: 'AbstractAgent', mode: StatModes = None, *,
-                 cache_lifespan: float = 120.0) -> None:
+                 cache_lifespan: int = 120) -> None:
         self.agent = agent
         self.mode = mode if mode is not None else StatModes.get_preferred_mode()
         self.cache_lifespan = cache_lifespan
@@ -273,16 +278,19 @@ class StatContext:
         Intended to be used by the agent.
         """
         async with self._lock:
-            # gather node/device metrics from compute plugins
+            # Here we use asyncio.gather() instead of aiotools.TaskGroup
+            # to keep methods of other plugins running when a plugin raises an error
+            # instead of cancelling them.
             _tasks = []
             for computer in self.agent.computers.values():
                 _tasks.append(computer.instance.gather_node_measures(self))
-            for node_measures in (await asyncio.gather(*_tasks, return_exceptions=True)):
-                if isinstance(node_measures, Exception):
-                    log.error('gather_node_measures error',
-                              exc_info=node_measures)
+            results = await asyncio.gather(*_tasks, return_exceptions=True)
+            for result in results:
+                if isinstance(result, Exception):
+                    log.error('collect_node_stat(): gather_node_measures() error',
+                              exc_info=result)
                     continue
-                for node_measure in node_measures:
+                for node_measure in result:
                     metric_key = node_measure.key
                     # update node metric
                     if metric_key not in self.node_metrics:
@@ -374,21 +382,25 @@ class StatContext:
                       self.agent.local_config['agent']['id'], redis_agent_updates['node'])
         serialized_agent_updates = msgpack.packb(redis_agent_updates)
 
-        def _pipe_builder():
-            pipe = self.agent.redis_stat_pool.pipeline()
-            pipe.set(self.agent.local_config['agent']['id'], serialized_agent_updates)
-            pipe.expire(self.agent.local_config['agent']['id'], self.cache_lifespan)
-            for kernel_id, metrics in self.kernel_metrics.items():
-                serialized_metrics = {
-                    key: obj.to_serializable_dict()
-                    for key, obj in metrics.items()
-                }
-                pipe.set(str(kernel_id), msgpack.packb(serialized_metrics))
-                pipe.expire(str(kernel_id), self.cache_lifespan)
-            return pipe
-        await redis.execute_with_retries(_pipe_builder)
+        async def _pipe_builder(r: aioredis.Redis):
+            async with r.pipeline() as pipe:
+                pipe.set(self.agent.local_config['agent']['id'], serialized_agent_updates)
+                pipe.expire(self.agent.local_config['agent']['id'], self.cache_lifespan)
+                for kernel_id, metrics in self.kernel_metrics.items():
+                    serialized_metrics = {
+                        key: obj.to_serializable_dict()
+                        for key, obj in metrics.items()
+                    }
+                    pipe.set(str(kernel_id), msgpack.packb(serialized_metrics))
+                    pipe.expire(str(kernel_id), self.cache_lifespan)
+                await pipe.execute()
 
-    async def collect_container_stat(self, container_id: ContainerId) -> Mapping[MetricKey, Metric]:
+        await redis.execute(self.agent.redis_stat_pool, _pipe_builder)
+
+    async def collect_container_stat(
+        self,
+        container_ids: Sequence[ContainerId],
+    ) -> Mapping[MetricKey, Metric]:
         """
         Collect the per-container statistics only,
 
@@ -399,22 +411,25 @@ class StatContext:
             for kid, info in self.agent.kernel_registry.items():
                 cid = info['container_id']
                 kernel_id_map[ContainerId(cid)] = kid
-
+            # Here we use asyncio.gather() instead of aiotools.TaskGroup
+            # to keep methods of other plugins running when a plugin raises an error
+            # instead of cancelling them.
             _tasks = []
             kernel_id = None
             for computer in self.agent.computers.values():
-                _tasks.append(
-                    computer.instance.gather_container_measures(self, [container_id]))
-            for ctnr_measures in (await asyncio.gather(*_tasks, return_exceptions=True)):
-                if isinstance(ctnr_measures, Exception):
-                    log.error('gather_cotnainer_measures error',
-                              exc_info=ctnr_measures)
+                _tasks.append(asyncio.create_task(
+                    computer.instance.gather_container_measures(self, container_ids)
+                ))
+            results = await asyncio.gather(*_tasks, return_exceptions=True)
+            for result in results:
+                if isinstance(result, Exception):
+                    log.error('collect_container_stat(): gather_container_measures() error',
+                              exc_info=result)
                     continue
-                for ctnr_measure in ctnr_measures:
+                for ctnr_measure in result:
                     metric_key = ctnr_measure.key
                     # update per-container metric
                     for cid, measure in ctnr_measure.per_container.items():
-                        assert cid == container_id
                         try:
                             kernel_id = kernel_id_map[cid]
                         except KeyError:
@@ -445,11 +460,12 @@ class StatContext:
                           kernel_id, serializable_metrics)
             serialized_metrics = msgpack.packb(serializable_metrics)
 
-            def _pipe_builder():
-                pipe = self.agent.redis_stat_pool.pipeline()
-                pipe.set(str(kernel_id), serialized_metrics)
-                pipe.expire(str(kernel_id), self.cache_lifespan)
-                return pipe
-            await redis.execute_with_retries(_pipe_builder)
+            async def _pipe_builder(r: aioredis.Redis):
+                async with r.pipeline() as pipe:
+                    pipe.set(str(kernel_id), serialized_metrics)
+                    pipe.expire(str(kernel_id), self.cache_lifespan)
+                    await pipe.execute()
+
+            await redis.execute(self.agent.redis_stat_pool, _pipe_builder)
             return metrics
         return {}
