@@ -133,8 +133,7 @@ async def handle_fstab_detail(request: web.Request) -> web.Response:
     log.info('HANDLE_FSTAB_DETAIL')
     params = request.query
     fstab_path = params.get('fstab_path', '/etc/fstab')
-    # FIXME: Remove ignore if https://github.com/python/typeshed/pull/4650 is released
-    async with aiofiles.open(fstab_path, mode='r') as fp:  # type: ignore
+    async with aiofiles.open(fstab_path, mode='r') as fp:
         content = await fp.read()
         return web.Response(text=content)
 
@@ -147,8 +146,7 @@ async def handle_list_mounts(request: web.Request) -> web.Response:
         mount_prefix = '/mnt'
     mounts = set()
     for p in Path(mount_prefix).iterdir():
-        # TODO: Change os.path.ismount to p.is_mount if Python 3.7 is supported.
-        if p.is_dir() and os.path.ismount(str(p)):
+        if p.is_dir() and p.is_mount():
             mounts.add(str(p))
     return web.json_response(sorted(mounts))
 
