@@ -41,24 +41,24 @@ class Runner(BaseRunner):
     }
     jupyter_kspec_name = 'python'
 
-    def __init__(self):
-        super().__init__()
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
         self.input_queue = None
         self.output_queue = None
 
     async def init_with_loop(self):
-        self.input_queue = janus.Queue(loop=self.loop)
-        self.output_queue = janus.Queue(loop=self.loop)
+        self.input_queue = janus.Queue()
+        self.output_queue = janus.Queue()
 
         # We have interactive input functionality for query mode!
-        self._user_input_queue = janus.Queue(loop=self.loop)
+        self._user_input_queue = janus.Queue()
         self.user_input_queue = self._user_input_queue.async_q
 
         # Get USER_SITE for runtime python.
         cmd = [self.runtime_path, *DEFAULT_PYFLAGS,
                '-c', 'import site; print(site.USER_SITE)']
         proc = await asyncio.create_subprocess_exec(
-            *cmd, env=self.child_env,
+            *map(str, cmd), env=self.child_env,
             stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)
         stdout, _ = await proc.communicate()
         user_site = stdout.decode('utf8').strip()
@@ -74,8 +74,8 @@ class Runner(BaseRunner):
     async def build_heuristic(self) -> int:
         if Path('setup.py').is_file():
             cmd = [
-                self.runtime_path, *DEFAULT_PYFLAGS,
-                '-m', 'pip', 'install', '-e', '.',
+                str(self.runtime_path), *DEFAULT_PYFLAGS,
+                '-m', 'pip', 'install', '--user', '-e', '.',
             ]
             return await self.run_subproc(cmd)
         else:
@@ -85,7 +85,7 @@ class Runner(BaseRunner):
     async def execute_heuristic(self) -> int:
         if Path('main.py').is_file():
             cmd = [
-                self.runtime_path, *DEFAULT_PYFLAGS,
+                str(self.runtime_path), *DEFAULT_PYFLAGS,
                 'main.py',
             ]
             return await self.run_subproc(cmd, batch=True)
@@ -128,6 +128,11 @@ class Runner(BaseRunner):
                 '--port', str(service_info['port']),
                 '--debugger_port', '6064',  # used by in-container TensorFlow
             ], {}
+        elif service_info['name'] == 'spectravis':
+            return [
+                self.runtime_path, '-m', 'http.server',
+                '8000',
+            ], {}, '/home/work/spectravis'
         elif service_info['name'] == 'sftp':
             return [
                 self.runtime_path, '-m', 'sftpserver', '--port', str(service_info['port'])
